REBOL [
	Title: "REBOL Mezzanine Functions: Function"
	Version: 2.7.8
	Rights: "Copyright REBOL Technologies 2009. All rights reserved."
	Home: http://www.rebol.com
	Date: 26-Mar-2010

	; You are free to use, modify, and distribute this file as long as the
	; above header, copyright, and this entire comment remains intact.
	; This software is provided "as is" without warranties of any kind.
	; In no event shall REBOL Technologies or source contributors be liable
	; for any damages of any kind, even if advised of the possibility of such
	; damage. See license for more information.

	; Please help us to improve this software by contributing changes and
	; fixes. See http://www.rebol.com/support.html for details.

	; Portions of this code are licensed from the R2/Forward project.
	; You are free to use, modify, and distribute this software with any
	; REBOL Technologies products as long as the above header, copyright,
	; and this comment remain intact. This software is provided "as is"
	; and without warranties of any kind. In no event shall the owners or
	; contributors be liable for any damages of any kind, even if advised
	; of the possibility of such damage. See license for more information.

	; Please help us to improve this software by contributing changes and
	; fixes via http://www.rebol.com/feedback.html - Thanks!
]

funct: func [
	"Defines a function with all set-words as locals."
	[catch]
	spec [block!] "Help string (opt) followed by arg words (and opt type and string)"
	body [block!] "The body block of the function"
	/with "Define or use a persistent object (self)"
	object [object! block!] "The object or spec"
	/extern words [block!] "These words are not local"
	/local r ws wb a
][
	spec: copy/deep spec
	body: copy/deep body
	; Get the words in the spec (ws) as word! so we can screen them out later
	ws: make block! length? spec
	parse spec [any [
		set-word! | set a any-word! (insert tail ws to-word a) | skip
	]]
	; Build the object if need be and screen out its words too
	if with [
		unless object? object [object: make object! object]
		bind body object  ; Bind any object words found in the body
		insert tail ws first object ; first used since 'self should be included
	]
	; Screen out the external words too
	insert tail ws words
	; Get any set-words in the code block as words (wb)
	wb: make block! 12  ; This should be a reasonable default
	parse body r: [any [
		set a set-word! (insert tail wb to-word a) |
		hash! | into r | skip
	]]
	; Remove the ws words from wb and add the rest to the spec as locals
	unless empty? wb: exclude wb ws [
		remove find wb 'local
		unless find spec /local [insert tail spec /local]
		insert tail spec wb
	]
	throw-on-error [make function! spec body]
]
; Note: The set-word! collection and spec word screening is native in R3.
; All new code written by BrianH based on an R3 version by BrianH and Carl.

closure: func [
	"Defines a closure function."
	[catch]
	spec [block!] "Help string (opt) followed by arg words (and opt type and string)"
	body [block!] "The body block of the function"
	/local spc bdy word
] [
	spc: make block! 1 + (2 * length? spec) ; 2 * in case arguments not typed
	insert/only spc [throw]
	bdy: make block! 5 + length? spec
	insert bdy reduce [:do :make :function! spc body]
	parse spec [any [
		set-word! | set word any-word! (
			insert tail bdy to word! :word
			insert tail spc to get-word! :word
			insert/only tail spc [any-type!]
		) | skip
	]]
	throw-on-error [make function! spec bdy]
]
; Note: Adapted from code by Ladislav Mecir, used with permission.
;   This uses the R2-style get-word! parameter word! retrieval.
;   The :do :make :function! signiature is checked in CLOSURE? below.

; R3 has TO-FUNCTION, so now R2 does too
to-function: func ["Converts to function! value." value [block!]] [
	make function! pick value 1 pick value 2 
] ; Should be close enough to the R3 version.

; Fake closure! with function!
closure!: :function!  ; Doesn't work in function specs, TYPE?
closure?: func ["Returns TRUE if it is this type." value [any-type!]] [
	all [
		function? get/any 'value
		value: second :value  ; Not doing BODY-OF here to lower overhead
		same? pick value 1 :do
		same? pick value 2 :make
		same? pick value 3 :function!
	] ; This should be good enough...
] ; Note: FUNCTION? will also succeed in R2
to-closure: func ["Converts to closure! value." value [block!]] [
	closure pick value 1 pick value 2 
] ; Should be close enough to the R3 version.

apply: func [
	"Apply a function to a reduced block of arguments."
	[throw]
	func [any-function!] "Function value to apply"
	block [block!] "Block of args, reduced first (unless /only)"
	/only "Use arg values as-is, do not reduce the block"
	/local words path todo noref value vars var
		v1 v2 v3 v4 v5 v6 v7 v8 v9 v10 v11 v12 v13 v14 v15 v16 v17 v18 v19 v20
][
	unless only [block: reduce block]
	words: first :func ; words-of
	vars: [ ; Used to special-case 'a and :a parameters (the quick way)
		v1 v2 v3 v4 v5 v6 v7 v8 v9 v10 v11 v12 v13 v14 v15 v16 v17 v18 v19 v20
	]
	path: to-path [func]  ; Even a one-element path works.
	todo: head insert/only make block! 1 + length? block path  ; Should be OK
	noref: false  ; True if the refinement isn't used
	while [not tail? words] [
		set/any 'value pick block 1  ; Safe for unset! and error!
		switch type?/word first words [
			word! [ ; Regular param, all values need evaluation blocked.
				unless noref [
					either word? get/any 'value [ ; Work around QUOTE limit
						insert tail todo to-lit-word get/any 'value
					] [
						insert/only insert tail todo 'quote get/any 'value
					]
				]
			]
			lit-word! [ ; Lit-word param, need to special-case get-words.
				unless noref [
					either get-word? get/any 'value [
						; Needs to be a get-word! assigned to a get-word! to get in
						either var: pick vars 1 [vars: next vars] [
							var: use [a] copy ['a]  ; else do it the slow way
						]
						set var value
						insert tail todo to-get-word var
					] [
						insert/only tail todo get/any 'value
					]
				]
			]
			get-word! [ ; Get-word param, need to special-case words.
				unless noref [
					either word? get/any 'value [
						; Needs to be a word! assigned to a word! to get in
						either var: pick vars 1 [vars: next vars] [
							var: use [a] copy ['a]  ; else do it the slow way
						]
						set var value
						insert tail todo var
					] [
						insert/only tail todo get/any 'value
					]
				]
			]
			refinement! [ ; Refinement, skip its associated params if not TRUE?
				unless noref: not get/any 'value [
					insert tail path to-word first words
				]
			]
		]
		words: next words
		block: next block
	]
	also do todo ( ; DO the built code, then cleanup memory references
		set [func words block path todo value vars var] set head vars none
	)
]
; Note: APPLY evaluates the arguments block (or not with /only), regardless of
; how parameters are specified for the function itself. This means that the
; normal evalation needs to be shortcircuited, so to speak - the REBOL code
; above does this. However, the R3 version is a native that is about as quick
; as a regular function call, while the R2 version is a bit slower.
; Even so, there are good reasons to call a function with APPLY:
; - APPLY does positional evaluation of refinements, so it is easier to write
;   wrapper functions that pass along refinements.
; - If you are calling function values that are passed as arguments to your
;   function, any lit-word! or get-word! parameters can be used to inject
;   code into your function, which can be a security hole. APPLY can make
;   functional-style REBOL code safer to write.
; Gotcha: Returning an error from the applied function triggers the error. It
; is not yet known how to fix this in R2 without modifying APPLY at runtime.
