REBOL [
	Title: "REBOL Mezzanine Functions: Reflection"
	Version: 2.7.8
	Rights: "Copyright REBOL Technologies 2009. All rights reserved."
	Home: http://www.rebol.com
	Date: 30-Jan-2010

	; You are free to use, modify, and distribute this file as long as the
	; above header, copyright, and this entire comment remains intact.
	; This software is provided "as is" without warranties of any kind.
	; In no event shall REBOL Technologies or source contributors be liable
	; for any damages of any kind, even if advised of the possibility of such
	; damage. See license for more information.

	; Please help us to improve this software by contributing changes and
	; fixes. See http://www.rebol.com/support.html for details.

	; Portions of this code are licensed from the R2/Forward project.
	; You are free to use, modify, and distribute this software with any
	; REBOL Technologies products as long as the above header, copyright,
	; and this comment remain intact. This software is provided "as is"
	; and without warranties of any kind. In no event shall the owners or
	; contributors be liable for any damages of any kind, even if advised
	; of the possibility of such damage. See license for more information.

	; Please help us to improve this software by contributing changes and
	; fixes via http://www.rebol.com/feedback.html - Thanks!
]

; Typeset spoofing, used by TYPES-OF

; Fake typeset! with block! of datatype!
typeset!: :block!
typeset?: func ["Returns TRUE if it is this type." value [any-type!]] [
	found? all [block? get/any 'value parse value [any datatype!]]
] ; Note: BLOCK? will also succeed in R2
to-typeset: funct [
	"Converts to typeset! value." [catch] value
][
	anytype: [
		none! logic! integer! decimal! money! char! pair! tuple! time! date!
		string! binary! file! email! url! tag! issue! bitset! image! block!
		paren! path! set-path! lit-path! datatype! word! set-word! get-word!
		lit-word! refinement! native! action! routine! op! function! object!
		error! port! event! struct! library! hash! list! symbol! unset!
	]
	anyblock: [
		block! paren! path! set-path! lit-path! hash! list!
	]
	anyfunction: [
		native! action! routine! op! function! ;rebcode!
	]
	anystring: [
		string! binary! file! email! url! tag! issue!
	]
	anyword: [
		word! set-word! get-word! lit-word! refinement!
	]
	series: [
		string! binary! file! email! url! tag! issue! image!
		block! paren! path! set-path! lit-path! hash! list!
	]
	number: [integer! decimal!]
	switch/default either datatype? :value [to-word value] [:value] [
		any-type! [reduce anytype]
		any-block! [reduce anyblock]
		any-function! [reduce anyfunction]
		any-string! [reduce anystring]
		any-word! [reduce anyword]
		series! [reduce series]
		number! [reduce number]
	] [
		switch/default type?/word :value [
			datatype! [reduce [value]]
			block! [
				parse value: copy value [any [
					datatype! | value: word! (
						change value throw-on-error [to-datatype first value]
					) | value: block! (
						value: change/part value to-typeset first value 1
					) :value |
					value: skip (
						throw-error 'script 'invalid-arg reduce [first value]
					)
				]]
				head value
			]
		] [throw-error 'script 'invalid-arg reduce [:value]] ; 'bad-make-arg in R3
	]
] ; MAKE or TO typeset! anything-else doesn't work - use TO-TYPESET.
; These blocks of datatypes can be used with FIND like R3 typesets.
; Note: You need to use FOUND? with FIND typeset in R2 to get the R3 result.
; R3 new types not included: percent! vector! get-path! map! typeset! rebcode!
;   command! closure! frame! module! task! gob! handle! utype!
; R2 pseudotypes special-cased:
;   any-block! any-function! any-string! any-type! any-word! series! number!
; R2 obsolete types included: hash! list! symbol!

; Fake any-path! typeset
any-path!: reduce [path! lit-path! set-path!]
any-path?: func [
	"Return TRUE if value is any type of path."
	value [any-type!]
][
	found? find any-path! type? get/any 'value
]

; Fake any-object! typeset
any-object!: reduce [object! error! port!] ; module! task!
any-object?: func [
	"Return TRUE if value is any type of object."
	value [any-type!]
][
	found? find any-object! type? get/any 'value
]

; Fake scalar! typeset
scalar!: reduce
	[integer! decimal! money! char! pair! tuple! time!]
scalar?: func [
	"Return TRUE if value is any type of scalar."
	value [any-type!]
][
	found? find scalar! type? get/any 'value
]

; Fake immediate! and internal! typesets, for documentation
immediate!: reduce [
	none! logic! integer! decimal! money! char! pair! tuple! time! date!
	datatype! word! set-word! get-word! lit-word! refinement! event!
] ; percent! typeset!
internal!: reduce [end! unset! symbol!] ; frame! handle!


; Reflection functions (reverse implementation compared to R3)

reflect: func [
	"Returns definition-related details about a value."
	[catch]
	value [any-type!]
	field [word!] "Such as: spec body words values title etc."
][throw-on-error [ ; In R3 the *-OF functions redirect to REFLECT - vice-versa in R2.
	switch/default field [
		spec [spec-of :value]
		body [body-of :value]
		words [words-of :value]
		values [values-of :value]
		types [types-of :value]
		title [title-of :value]
	] [
		cause-error 'script 'invalid-arg field
	]
]]
; Note: Native (action!) in R3, and the *-OF functions below are wrappers.

spec-of: func [
	"Returns a copy of the spec of a function."
	value
][
	case [
		object? :value [none]
		any-function? :value [copy/deep third :value]
		'else [cause-error 'script 'cannot-use reduce ['reflect type? :value]]
	]
]

body-of: func [
	"Returns a copy of the body of a function or object."
	value
][
	case [
		object? :value [third :value]
		function? :value [copy/deep second :value] ; Note: Still bound!
		any-function? :value [none] ; none if native
		'else [cause-error 'script 'cannot-use reduce ['reflect type? :value]]
	]
]

words-of: func [
	"Returns a copy of the words of a function or object."
	value
][
	case [
		object? :value [bind remove first :value :value] ; removes 'self
		any-function? :value [first :value]
		'else [cause-error 'script 'cannot-use reduce ['reflect type? :value]]
	]
]

values-of: func [
	"Returns a copy of the values of an object."
	value
][
	case [
		object? :value [remove second :value] ; removes :self
		any-function? :value [none]
		'else [cause-error 'script 'cannot-use reduce ['reflect type? :value]]
	]
]

types-of: funct [
	"Returns a copy of the types of a function."
	value
] compose/only [ ; Returns types as blocks of datatype! values
	valtype: (
		head remove find head remove find to-typeset any-type! error! unset!
	) ; R2 excludes error! here where R3 doesn't
	reftype: [none! logic!]
	case [
		object? :value [cause-error 'script 'invalid-arg 'types]
		any-function? :value [
			result: copy []
			types: none
			parse third :value [any [to any-word! [
				[word! | lit-word! | get-word!] [
					set types block! (
						types: copy types
						while [not tail? types] [
							switch/default to-word first types [
								any-type! [types: change/part types to-typeset any-type! 1]
								any-block! [types: change/part types to-typeset any-block! 1]
								any-function! [types: change/part types to-typeset any-function! 1]
								any-string! [types: change/part types to-typeset any-string! 1]
								any-word! [types: change/part types to-typeset any-word! 1]
								series! [types: change/part types to-typeset series! 1]
								number! [types: change/part types to-typeset number! 1]
							] [
								types: next types
							]
						]
						insert/only tail result head types
					) | (
						insert/only tail result reduce valtype
					)
				] |
				refinement! (
					insert/only tail result reduce reftype
				) |
				skip ; I have no idea how set-words in the spec are to be treated
			]]]
			result
		]
		'else [cause-error 'script 'cannot-use reduce ['reflect type? :value]]
	]
]
; Note: R3 doesn't exclude error! from the default parameter types, but R2 does.
; See the typeset! spoofing for details about the return values.

title-of: func [
	"Returns a copy of the title of a function."
	value
][
	case [
		object? :value [cause-error 'script 'invalid-arg 'title]
		any-function? :value [if string? value: pick third :value 1 [copy value]]
		'else [cause-error 'script 'cannot-use reduce ['reflect type? :value]]
	]
]


; ASSERT, uses typeset spoofing too

assert: func [
	"Assert that condition is true, else throw an assertion error."
	[catch throw]
	conditions [block!]
	/type "Safely check datatypes of variables (words)" ; not paths
	/local w t
][throw-on-error [
	either type [
		parse conditions [any [
			[set w word! | set w skip (
				cause-error 'script 'invalid-arg type? get/any 'w
			)]
			[set t [block! | word!] (
				unless find to-typeset t type? get/any w [
					make error! join "datatype assertion failed for: " w
				]
			) | set t skip (
				cause-error 'script 'invalid-arg type? get/any 't
			)]
		]]
	][
		any [
			all conditions
			make error! join "assertion failed for: " mold conditions
		]
	]
]]
; This is a relatively slow, loose approximation of a native in R3.
; ASSERT/type of paths doesn't work yet, since GET doesn't either.
